import json
from copy import deepcopy
from hashlib import sha1

from lxml.etree import SubElement  # nosec - see 'lxml` package (bandit)' section in README

from bas_metadata_library.standards.iso_19115_common import MetadataRecordElement
from bas_metadata_library.standards.iso_19115_common.common_elements import (
    Format,
    OnlineResource,
    ResponsibleParty,
)
from bas_metadata_library.standards.iso_19115_common.utils import (
    condense_distribution_distributors,
    flatten_distribution_distributors,
    format_distribution_option_consistently,
    format_numbers_consistently,
)


class DataDistribution(MetadataRecordElement):
    def make_config(self) -> list:
        _ = []

        distributions_length = int(
            self.record.xpath(
                f"count({self.xpath}/gmd:distributionInfo/gmd:MD_Distribution/gmd:distributor)",
                namespaces=self.ns.nsmap(),
            )
        )
        for distribution_index in range(1, distributions_length + 1):
            distribution = Distribution(
                record=self.record,
                attributes=self.attributes,
                xpath=f"{self.xpath}/gmd:distributionInfo/gmd:MD_Distribution/gmd:distributor[{distribution_index}]/"
                f"gmd:MD_Distributor",
            )
            _distribution = distribution.make_config()
            if bool(_distribution):
                _.append(_distribution)

        # distributions are grouped by distributor in ISO but should not be in configs
        return flatten_distribution_distributors(distributions=_)

    def make_element(self) -> None:
        data_distribution_wrapper = SubElement(self.record, f"{{{self.ns.gmd}}}distributionInfo")
        data_distribution_element = SubElement(data_distribution_wrapper, f"{{{self.ns.gmd}}}MD_Distribution")

        # distributions are grouped by distributor in ISO but aren't in configs
        # noinspection PyTypeChecker
        distributions = condense_distribution_distributors(distributions=self.element_attributes)

        for distribution_config in distributions:
            distribution_element = Distribution(
                record=self.record,
                attributes=self.attributes,
                parent_element=data_distribution_element,
                element_attributes=distribution_config,
            )
            distribution_element.make_element()


class Distribution(MetadataRecordElement):
    def make_config(self) -> dict:
        _ = {}

        distributor = Distributor(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:distributorContact",
        )
        _distributor = distributor.make_config()
        if bool(_distributor):
            _["distributor"] = _distributor

        _distribution_formats = []
        distribution_formats_length = int(
            self.record.xpath(
                f"count({self.xpath}/gmd:distributorFormat)",
                namespaces=self.ns.nsmap(),
            )
        )
        for distribution_format_index in range(1, distribution_formats_length + 1):
            distribution_option_format = DistributorFormat(
                record=self.record,
                attributes=self.attributes,
                xpath=f"{self.xpath}/gmd:distributorFormat[{distribution_format_index}]",
            )
            _distribution_option_format = distribution_option_format.make_config()
            if bool(_distribution_option_format):
                _distribution_formats.append(_distribution_option_format)

        _transfer_options = []
        transfer_options_length = int(
            self.record.xpath(
                f"count({self.xpath}/gmd:distributorTransferOptions)",
                namespaces=self.ns.nsmap(),
            )
        )
        for transfer_option_index in range(1, transfer_options_length + 1):
            distribution_transfer_option = DistributorTransferOption(
                record=self.record,
                attributes=self.attributes,
                xpath=f"{self.xpath}/gmd:distributorTransferOptions[{transfer_option_index}]",
            )
            _distribution_transfer_option = distribution_transfer_option.make_config()
            if bool(_distribution_transfer_option):
                _transfer_options.append(_distribution_transfer_option)

        _["distribution_options"] = self._match_distribution_options(
            distribution_formats=_distribution_formats, transfer_options=_transfer_options
        )

        return _

    def make_element(self) -> None:
        distribution_container = SubElement(self.parent_element, f"{{{self.ns.gmd}}}distributor")
        distribution_element = SubElement(distribution_container, f"{{{self.ns.gmd}}}MD_Distributor")

        distributor = Distributor(
            record=self.record,
            attributes=self.attributes,
            parent_element=distribution_element,
            element_attributes=self.element_attributes["distributor"],
        )
        distributor.make_element()

        _formats: list[DistributorFormat] = []
        _transfer_options: list[DistributorTransferOption] = []

        for distribution_option in self.element_attributes["distribution_options"]:
            # ID elements are generated by hashing a JSON encoding of the entire distribution object,
            # see the README 'Automatic transfer option / format IDs' section for more information
            _distribution_option = format_distribution_option_consistently(distribution_option=distribution_option)
            # S324 warning is exempted as these hashes are not used for any security related purposes
            _id = sha1(json.dumps(_distribution_option).encode()).hexdigest()  # noqa: S324

            if "format" in distribution_option:
                distribution_format = DistributorFormat(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=distribution_element,
                    element_attributes={**distribution_option["format"], "_id": _id},
                )
                _formats.append(distribution_format)

            if "transfer_option" in distribution_option:
                transfer_option = DistributorTransferOption(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=distribution_element,
                    element_attributes={**distribution_option["transfer_option"], "_id": _id},
                )
                _transfer_options.append(transfer_option)

        for _format in _formats:
            _format.make_element()
        for _transfer_option in _transfer_options:
            _transfer_option.make_element()

    @staticmethod
    def _match_distribution_options(distribution_formats: list[dict], transfer_options: list[dict]) -> list[dict]:  # noqa: C901 see uk-pdc/metadata-infrastructure/metadata-library#175 for more information
        """
        Utility method to match distribution formats and transfer options into complete or partial distribution objects.

        See the README 'Automatic transfer option / format IDs' section for more information on the background to this
        feature and why it's needed.

        Note: there is almost certainly a more elegant/efficient way to do this process. Alternative/improved
        implementations are definitely welcome as contributions.

        :param distribution_formats: list of distribution format objects
        :param transfer_options: list of transfer option objects
        :return list of complete or partial distribution option objects
        """
        distribution_options = []
        _distribution_formats = {}
        _transfer_options = {}
        _unmatched_distribution_formats = []
        _unmatched_transfer_options = []
        _matched_distribution_formats = []
        _matched_transfer_options = []

        # index options by ID or add to unmatched list
        for distribution_format in distribution_formats:
            if "_id" in distribution_format:
                _id = distribution_format["_id"]
                del distribution_format["_id"]
                _distribution_formats[_id] = distribution_format
                continue
            _unmatched_distribution_formats.append(distribution_format)
        for transfer_option in transfer_options:
            if "_id" in transfer_option:
                _id = transfer_option["_id"]
                del transfer_option["_id"]
                _transfer_options[_id] = transfer_option
                continue
            _unmatched_transfer_options.append(transfer_option)

        # try to match up items or add to unmatched list
        if len(_distribution_formats) >= len(_transfer_options):
            for fmt_id, distribution_format in _distribution_formats.items():
                if fmt_id not in _transfer_options:
                    _unmatched_distribution_formats.append(distribution_format)
                    continue

                distribution_options.append(
                    {"format": distribution_format, "transfer_option": _transfer_options[fmt_id]}
                )
                _matched_distribution_formats.append(fmt_id)
                _matched_transfer_options.append(fmt_id)
        if len(_transfer_options) > len(_distribution_formats):
            for tfo_id, transfer_option in _transfer_options.items():
                if tfo_id not in _distribution_formats:
                    _unmatched_transfer_options.append(transfer_option)
                    continue

                distribution_options.append(
                    {"format": _distribution_formats[tfo_id], "transfer_option": transfer_option}
                )
                _matched_transfer_options.append(tfo_id)
                _matched_distribution_formats.append(tfo_id)

        # add all unmatched items as non-complete distribution options
        for unmatched_distribution_format in _unmatched_distribution_formats:
            distribution_options.append({"format": unmatched_distribution_format})
        for unmatched_transfer_options in _unmatched_transfer_options:
            distribution_options.append({"transfer_option": unmatched_transfer_options})

        return distribution_options


class Distributor(MetadataRecordElement):
    def make_config(self) -> dict:
        responsible_party = ResponsibleParty(record=self.record, attributes=self.attributes, xpath=self.xpath)
        return responsible_party.make_config()

    def make_element(self) -> None:
        distributor_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}distributorContact")

        # roles need to looped through, but will always be 'distributor' only for distributors
        self.element_attributes = deepcopy(self.element_attributes)
        self.element_attributes["role"] = self.element_attributes["role"][0]

        responsible_party = ResponsibleParty(
            record=self.record,
            attributes=self.attributes,
            parent_element=distributor_element,
            element_attributes=self.element_attributes,
        )
        responsible_party.make_element()


class DistributorFormat(MetadataRecordElement):
    def make_config(self) -> dict:
        distributor_format = Format(record=self.record, attributes=self.attributes, xpath=self.xpath)
        return distributor_format.make_config()

    def make_element(self) -> None:
        distributor_format_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}distributorFormat")

        distributor_format = Format(
            record=self.record,
            attributes=self.attributes,
            parent_element=distributor_format_element,
            element_attributes=self.element_attributes,
        )
        distributor_format.make_element()


class DistributorTransferOption(MetadataRecordElement):
    def make_config(self) -> dict:
        _ = {}

        format_id = self.record.xpath(f"{self.xpath}/gmd:MD_DigitalTransferOptions/@id", namespaces=self.ns.nsmap())
        if len(format_id) == 1:
            _id: str = format_id[0].replace("bml-", "")
            _id = _id.replace("-tfo", "")
            _["_id"] = _id

        size_unit = self.record.xpath(
            f"{self.xpath}/gmd:MD_DigitalTransferOptions/gmd:unitsOfDistribution/gco:CharacterString/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(size_unit) == 1:
            if "size" not in _:
                _["size"] = {}
            _["size"]["unit"] = size_unit[0]

        size_magnitude = self.record.xpath(
            f"{self.xpath}/gmd:MD_DigitalTransferOptions/gmd:transferSize/gco:Real/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(size_magnitude) == 1:
            if "size" not in _:
                _["size"] = {}
            _["size"]["magnitude"] = format_numbers_consistently(size_magnitude[0])

        online_resource = OnlineResource(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:MD_DigitalTransferOptions/gmd:onLine",
        )
        _online_resource = online_resource.make_config()
        if bool(_online_resource):
            _["online_resource"] = _online_resource

        if list(_.keys()) == ["_id"]:
            _ = {}
        return _

    def make_element(self) -> None:
        transfer_options_container = SubElement(self.parent_element, f"{{{self.ns.gmd}}}distributorTransferOptions")
        transfer_options_wrapper = SubElement(
            transfer_options_container,
            f"{{{self.ns.gmd}}}MD_DigitalTransferOptions",
            attrib={"id": f"bml-{self.element_attributes['_id']}-tfo"},
        )

        if "size" in self.element_attributes:
            if "unit" in self.element_attributes["size"]:
                transfer_size_unit_element = SubElement(
                    transfer_options_wrapper, f"{{{self.ns.gmd}}}unitsOfDistribution"
                )
                transfer_size_unit_value = SubElement(transfer_size_unit_element, f"{{{self.ns.gco}}}CharacterString")
                transfer_size_unit_value.text = self.element_attributes["size"]["unit"]
            if "magnitude" in self.element_attributes["size"]:
                transfer_size_magnitude_element = SubElement(transfer_options_wrapper, f"{{{self.ns.gmd}}}transferSize")
                transfer_size_magnitude_value = SubElement(transfer_size_magnitude_element, f"{{{self.ns.gco}}}Real")
                transfer_size_magnitude_value.text = str(
                    format_numbers_consistently(self.element_attributes["size"]["magnitude"])
                )

        transfer_options_element = SubElement(transfer_options_wrapper, f"{{{self.ns.gmd}}}onLine")
        online_resource = OnlineResource(
            record=self.record,
            attributes=self.attributes,
            parent_element=transfer_options_element,
            element_attributes=self.element_attributes["online_resource"],
        )
        online_resource.make_element()



from lxml.etree import SubElement

from bas_metadata_library.standards.iso_19115_common import MetadataRecordElement
from bas_metadata_library.standards.iso_19115_common.base_elements import ScopeCode
from bas_metadata_library.standards.iso_19115_common.common_elements import Citation, ResponsibleParty
from bas_metadata_library.standards.iso_19115_common.utils import (
    condense_contacts_roles,
    decode_date_string,
    encode_date_string,
)


class DataQuality(MetadataRecordElement):
    """gmd:dataQualityInfo."""

    def make_config(self) -> dict:
        """Decode to Python."""
        _ = {}

        _domain_consistency = []
        domains_length = int(
            self.record.xpath(
                f"count({self.xpath}/gmd:dataQualityInfo/gmd:DQ_DataQuality/gmd:report/gmd:DQ_DomainConsistency)",
                namespaces=self.ns.nsmap(),
            )
        )
        for domain_index in range(1, domains_length + 1):
            domain_consistency = DomainConsistency(
                record=self.record,
                attributes=self.attributes,
                xpath=f"({self.xpath}/gmd:dataQualityInfo/gmd:DQ_DataQuality/gmd:report[gmd:DQ_DomainConsistency])[{domain_index}]",
            )
            _domain = domain_consistency.make_config()
            if bool(_domain):
                _domain_consistency.append(_domain)
        if len(_domain_consistency) > 0:
            _["domain_consistency"] = _domain_consistency

        lineage = Lineage(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:dataQualityInfo/gmd:DQ_DataQuality/gmd:lineage",
        )
        _lineage = lineage.make_config()
        if bool(_lineage):
            _["lineage"] = _lineage

        return _

    def make_element(self) -> None:
        """Encode as XML."""
        data_quality_wrapper = SubElement(self.record, f"{{{self.ns.gmd}}}dataQualityInfo")
        data_quality_element = SubElement(data_quality_wrapper, f"{{{self.ns.gmd}}}DQ_DataQuality")

        scope = Scope(record=self.record, attributes=self.attributes, parent_element=data_quality_element)
        scope.make_element()

        if "domain_consistency" in self.element_attributes["identification"]:
            for domain_consistency_attributes in self.element_attributes["identification"]["domain_consistency"]:
                domain_consistency = DomainConsistency(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=data_quality_element,
                    element_attributes=domain_consistency_attributes,
                )
                domain_consistency.make_element()

        if "lineage" in self.element_attributes["identification"]:
            lineage = Lineage(
                record=self.record,
                attributes=self.attributes,
                parent_element=data_quality_element,
                element_attributes=self.element_attributes["identification"]["lineage"],
            )
            lineage.make_element()


class Scope(MetadataRecordElement):
    """gmd:scope."""

    def make_element(self) -> None:
        """Encode as XML."""
        scope_wrapper = SubElement(self.parent_element, f"{{{self.ns.gmd}}}scope")
        scope_element = SubElement(scope_wrapper, f"{{{self.ns.gmd}}}DQ_Scope")

        scope_code = ScopeCode(record=self.record, attributes=self.attributes, parent_element=scope_element)
        scope_code.make_element()


class Lineage(MetadataRecordElement):
    """gmd:lineage."""

    def make_config(self) -> dict[str, str]:
        """Decode to Python."""
        _ = {}

        statement_value = self.record.xpath(
            f"{self.xpath}/gmd:LI_Lineage/gmd:statement/gco:CharacterString/text()", namespaces=self.ns.nsmap()
        )
        if len(statement_value) == 1:
            _["statement"] = statement_value[0]

        _process_steps = []
        steps_length = int(
            self.record.xpath(f"count({self.xpath}/gmd:LI_Lineage/gmd:processStep)", namespaces=self.ns.nsmap())
        )
        for step_index in range(1, steps_length + 1):
            process_step = ProcessStep(
                record=self.record,
                attributes=self.attributes,
                xpath=f"({self.xpath}/gmd:LI_Lineage/gmd:processStep)[{step_index}]",
            )
            _process_step = process_step.make_config()
            if bool(_process_step):
                _process_steps.append(_process_step)
        if len(_process_steps) > 0:
            _["process_steps"] = _process_steps

        _sources = []
        sources_length = int(
            self.record.xpath(f"count({self.xpath}/gmd:LI_Lineage/gmd:source)", namespaces=self.ns.nsmap())
        )
        for source_index in range(1, sources_length + 1):
            source = Source(
                record=self.record,
                attributes=self.attributes,
                xpath=f"({self.xpath}/gmd:LI_Lineage/gmd:source)[{source_index}]",
            )
            _source = source.make_config()
            if bool(_source):
                _sources.append(_source)
        if len(_sources) > 0:
            _["sources"] = _sources

        return _

    def make_element(self) -> None:
        """Encode as XML."""
        lineage_wrapper = SubElement(self.parent_element, f"{{{self.ns.gmd}}}lineage")
        lineage_element = SubElement(lineage_wrapper, f"{{{self.ns.gmd}}}LI_Lineage")

        if "statement" in self.element_attributes:
            statement_element = SubElement(lineage_element, f"{{{self.ns.gmd}}}statement")
            statement_value = SubElement(statement_element, f"{{{self.ns.gco}}}CharacterString")
            statement_value.text = self.element_attributes["statement"]

        if "process_steps" in self.element_attributes:
            for process_step_attributes in self.element_attributes["process_steps"]:
                process_step_wrapper = SubElement(lineage_element, f"{{{self.ns.gmd}}}processStep")

                process_step = ProcessStep(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=process_step_wrapper,
                    element_attributes=process_step_attributes,
                )
                process_step.make_element()

        if "sources" in self.element_attributes:
            for source_attributes in self.element_attributes["sources"]:
                source_wrapper = SubElement(lineage_element, f"{{{self.ns.gmd}}}source")

                source = Source(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=source_wrapper,
                    element_attributes=source_attributes,
                )
                source.make_element()


class ProcessStep(MetadataRecordElement):
    """gmd:LI_ProcessStep."""

    def make_config(self) -> dict[str, str]:
        """Decode to Python."""
        _ = {}

        description_value = self.record.xpath(
            f"{self.xpath}/gmd:LI_ProcessStep/gmd:description/gco:CharacterString/text()", namespaces=self.ns.nsmap()
        )
        if len(description_value) == 1:
            _["description"] = description_value[0]

        rationale_value = self.record.xpath(
            f"{self.xpath}/gmd:LI_ProcessStep/gmd:rationale/gco:CharacterString/text()", namespaces=self.ns.nsmap()
        )
        if len(rationale_value) == 1:
            _["rationale"] = rationale_value[0]

        date_value = self.record.xpath(
            f"{self.xpath}/gmd:LI_ProcessStep/gmd:dateTime/gco:DateTime/text()", namespaces=self.ns.nsmap()
        )
        if len(date_value) == 1:
            _["date"] = decode_date_string(date_datetime=date_value[0])["date"]  # unwrap datetime value

        _processors = []
        processors_length = int(
            self.record.xpath(f"count({self.xpath}/gmd:LI_ProcessStep/gmd:processor)", namespaces=self.ns.nsmap())
        )
        for processor_index in range(1, processors_length + 1):
            processor = Processor(
                record=self.record,
                attributes=self.attributes,
                xpath=f"({self.xpath}/gmd:LI_ProcessStep/gmd:processor)[{processor_index}]",
            )
            _processor = processor.make_config()
            if bool(_processor):
                _processors.append(_processor)
        if len(_processors) > 0:
            _processors = condense_contacts_roles(contacts=_processors)
            _["processors"] = _processors

        _sources = []
        sources_length = int(
            self.record.xpath(f"count({self.xpath}/gmd:LI_ProcessStep/gmd:source)", namespaces=self.ns.nsmap())
        )
        for source_index in range(1, sources_length + 1):
            source = Source(
                record=self.record,
                attributes=self.attributes,
                xpath=f"({self.xpath}/gmd:LI_ProcessStep/gmd:source)[{source_index}]",
            )
            _source = source.make_config()
            if bool(_source):
                _sources.append(_source)
        if len(_sources) > 0:
            _["sources"] = _sources

        return _

    def make_element(self) -> None:
        """Encode as XML."""
        process_step_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}LI_ProcessStep")

        description_element = SubElement(process_step_element, f"{{{self.ns.gmd}}}description")
        description_value = SubElement(description_element, f"{{{self.ns.gco}}}CharacterString")
        description_value.text = self.element_attributes["description"]

        if "rationale" in self.element_attributes:
            rational_element = SubElement(process_step_element, f"{{{self.ns.gmd}}}rationale")
            rational_value = SubElement(rational_element, f"{{{self.ns.gco}}}CharacterString")
            rational_value.text = self.element_attributes["rationale"]

        if "date" in self.element_attributes:
            date_element = SubElement(process_step_element, f"{{{self.ns.gmd}}}dateTime")
            date_value = SubElement(date_element, f"{{{self.ns.gco}}}DateTime")
            date_value.text = encode_date_string(self.element_attributes["date"])

        if "processors" in self.element_attributes:
            for processor_attributes in self.element_attributes["processors"]:
                for role in processor_attributes["role"]:
                    _processor_attributes = processor_attributes.copy()
                    _processor_attributes["role"] = role

                    processor = Processor(
                        record=self.record,
                        attributes=self.attributes,
                        parent_element=process_step_element,
                        element_attributes=_processor_attributes,
                    )
                    processor.make_element()

        if "sources" in self.element_attributes:
            for source_attributes in self.element_attributes["sources"]:
                source_wrapper = SubElement(process_step_element, f"{{{self.ns.gmd}}}source")

                source = Source(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=source_wrapper,
                    element_attributes=source_attributes,
                )
                source.make_element()


class Processor(MetadataRecordElement):
    """gmd:processor."""

    def make_config(self) -> dict:
        """Decode to Python."""
        responsible_party = ResponsibleParty(record=self.record, attributes=self.attributes, xpath=self.xpath)
        return responsible_party.make_config()

    def make_element(self) -> None:
        """Encode as XML."""
        processor_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}processor")

        responsible_party = ResponsibleParty(
            record=self.record,
            attributes=self.attributes,
            parent_element=processor_element,
            element_attributes=self.element_attributes,
        )
        responsible_party.make_element()


class Source(MetadataRecordElement):
    """gmd:source."""

    def make_config(self) -> dict:
        """Decode to Python."""
        _ = {}

        description_value = self.record.xpath(
            f"{self.xpath}/gmd:LI_Source/gmd:description/gco:CharacterString/text()", namespaces=self.ns.nsmap()
        )
        if len(description_value) == 1:
            _["description"] = description_value[0]

        citation = Citation(
            record=self.record, attributes=self.attributes, xpath=f"({self.xpath}/gmd:LI_Source/gmd:sourceCitation)"
        )
        _citation = citation.make_config()
        if bool(_citation):
            _.update(**_citation)

        _source_steps = []
        steps_length = int(
            self.record.xpath(f"count({self.xpath}/gmd:LI_Source/gmd:sourceStep)", namespaces=self.ns.nsmap())
        )
        for step_index in range(1, steps_length + 1):
            source_step = ProcessStep(
                record=self.record,
                attributes=self.attributes,
                xpath=f"({self.xpath}/gmd:LI_Source/gmd:sourceStep)[{step_index}]",
            )
            _source_step = source_step.make_config()
            if bool(_source_step):
                _source_steps.append(_source_step)
        if len(_source_steps) > 0:
            _["source_steps"] = _source_steps

        return _

    def make_element(self) -> None:
        """Encode to XML."""
        source_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}LI_Source")

        if "description" in self.element_attributes:
            description_element = SubElement(source_element, f"{{{self.ns.gmd}}}description")
            description_value = SubElement(description_element, f"{{{self.ns.gco}}}CharacterString")
            description_value.text = self.element_attributes["description"]

        # partial citation attributes
        if any(key in ["title", "dates", "edition", "identifiers", "contact"] for key in self.element_attributes):
            citation_element = SubElement(source_element, f"{{{self.ns.gmd}}}sourceCitation")
            citation = Citation(
                record=self.record,
                attributes=self.attributes,
                parent_element=citation_element,
                element_attributes=self.element_attributes,
            )
            citation.make_element()

        if "source_steps" in self.element_attributes:
            for step_attributes in self.element_attributes["source_steps"]:
                source_step_wrapper = SubElement(source_element, f"{{{self.ns.gmd}}}sourceStep")

                step = ProcessStep(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=source_step_wrapper,
                    element_attributes=step_attributes,
                )
                step.make_element()


class DomainConsistency(MetadataRecordElement):
    """gmd:DQ_DomainConsistency."""

    def make_config(self) -> dict:
        """Decode to Python."""
        _ = {}
        xpath = f"{self.xpath}/gmd:DQ_DomainConsistency/gmd:result/gmd:DQ_ConformanceResult"

        specification = Citation(record=self.record, attributes=self.attributes, xpath=f"({xpath}/gmd:specification)")
        _specification = specification.make_config()
        if bool(_specification):
            _["specification"] = _specification

        explanation_value = self.record.xpath(
            f"{xpath}/gmd:explanation/gco:CharacterString/text()", namespaces=self.ns.nsmap()
        )
        if len(explanation_value) == 1:
            _["explanation"] = explanation_value[0]

        result_value = self.record.xpath(f"{xpath}/gmd:pass/gco:Boolean/text()", namespaces=self.ns.nsmap())
        if len(result_value) == 1:
            _["result"] = bool(result_value[0])

        return _

    def make_element(self) -> None:
        """Encode as XML."""
        domain_conformance_wrapper = SubElement(self.parent_element, f"{{{self.ns.gmd}}}report")
        domain_conformance_element = SubElement(domain_conformance_wrapper, f"{{{self.ns.gmd}}}DQ_DomainConsistency")
        result_wrapper = SubElement(domain_conformance_element, f"{{{self.ns.gmd}}}result")
        result_element = SubElement(result_wrapper, f"{{{self.ns.gmd}}}DQ_ConformanceResult")

        specification_element = SubElement(result_element, f"{{{self.ns.gmd}}}specification")
        citation = Citation(
            record=self.record,
            attributes=self.attributes,
            parent_element=specification_element,
            element_attributes=self.element_attributes["specification"],
        )
        citation.make_element()

        explanation_element = SubElement(result_element, f"{{{self.ns.gmd}}}explanation")
        explanation_value = SubElement(explanation_element, f"{{{self.ns.gco}}}CharacterString")
        explanation_value.text = self.element_attributes["explanation"]

        pass_element = SubElement(result_element, f"{{{self.ns.gmd}}}pass")
        pass_value = SubElement(pass_element, f"{{{self.ns.gco}}}Boolean")
        pass_value.text = str(self.element_attributes["result"]).lower()



from __future__ import annotations

import json
from copy import deepcopy
from datetime import date, datetime
from itertools import groupby


def _sort_dict_by_keys(dictionary: dict) -> dict:
    """
    Utility method to recursively sort a dictionary by its keys.

    Keys are sorted alphabetically in ascending order.

    :param dictionary: input dictionary
    :return dictionary sorted by keys
    """
    return {k: _sort_dict_by_keys(v) if isinstance(v, dict) else v for k, v in sorted(dictionary.items())}


def _decode_date_properties(dictionary: dict, parent_keys: list[str] | None = None) -> dict:
    """
    Utility method to convert any date(time) values in a record configuration into a Python date(time) object.

    Using the `decode_date_string()` method (to handle partial dates).

    Date properties that are searched for:
    - metadata.date_stamp
    - *.dates.*
    - identification.extent.temporal.period.*

    E.g.
    - '2012-02-20' becomes `{'date': date(2012, 2, 20)}`
    - '2012-02' becomes `{'date': date(2012, 1, 1), 'date_precision': 'year'}`

    :param parent_keys: list of parent keys when current key/value has been accessed through recursion
    :return dictionary with parsed property values
    """
    if parent_keys is None:
        parent_keys = []

    for key, value in list(dictionary.items()):
        # check if value in dict is itself a dictionary, or a list of dictionaries, and recurse into if so
        if isinstance(value, list):
            for list_value in value:
                if isinstance(list_value, dict):
                    parent_keys.append(key)
                    _decode_date_properties(dictionary=list_value, parent_keys=parent_keys)
                    parent_keys.remove(key)
        elif isinstance(value, dict):
            parent_keys.append(key)
            _decode_date_properties(dictionary=value, parent_keys=parent_keys)
            parent_keys.remove(key)

        # check if key is a property that will contain a date(time) value
        elif isinstance(value, str) and "metadata" in parent_keys and key == "date_stamp":
            dictionary[key] = date.fromisoformat(value)
        elif (
            isinstance(value, str)
            and ("process_steps" in parent_keys or "source_steps" in parent_keys)
            and key == "date"
        ):
            dictionary[key] = decode_date_string(date_datetime=value)["date"]  # unwrap date value
        elif (
            isinstance(value, str)
            and "dates" in parent_keys
            or (
                isinstance(value, str)
                and "identification" in parent_keys
                and ("extent" in parent_keys or "extents" in parent_keys)
                and "temporal" in parent_keys
                and "period" in parent_keys
            )
        ):
            dictionary[key] = decode_date_string(date_datetime=value)

    return dictionary


def _encode_date_properties(dictionary: dict) -> dict:
    """
    Utility method to recursively convert any date values into a string value.

    Using the `encode_date_string()` utility method.

    Dates are represented as a dict with a date property, containing a date or datetime object, plus an optional date
    precision value. As dates are dicts, it's necessary for this method to check, when recusing through values, whether
    a value is a date dict, or a property dict. Otherwise, this method will produce incorrect results.

    :param dictionary: input dictionary
    :return dictionary with encoded property values
    """
    for key, value in list(dictionary.items()):
        if isinstance(value, list):
            # list of dictionaries, loop and recurse into
            for list_value in value:
                if isinstance(list_value, dict):
                    _encode_date_properties(dictionary=list_value)
        elif isinstance(value, dict) and list(value.keys()) == ["date"]:
            # date or datetime
            dictionary[key] = encode_date_string(date_datetime=value["date"])
        elif isinstance(value, dict) and {"date", "date_precision"}.issubset(set(value.keys())):
            # date or datetime with precision
            dictionary[key] = encode_date_string(date_datetime=value["date"], date_precision=value["date_precision"])
        elif isinstance(value, dict):
            # value is dictionary itself, recurse into
            _encode_date_properties(dictionary=value)
        elif isinstance(value, date) and key == "date_stamp":
            # metadata.date_stamp is always a date
            dictionary[key] = value.isoformat()
        elif isinstance(value, date) and key == "date":
            # lineage.process_step.date is always a datetime
            dictionary[key] = value.isoformat()

    return dictionary


def encode_date_string(date_datetime: date | datetime, date_precision: str | None = None) -> str:
    """
    Formats a python date or datetime object as an ISO 8601 date or datetime string representation.

    This method includes support for partial dates (year or year month) via an optional date precision element.
    When set the month and/or day elements in the Python date object are ignored when encoding as an ISO date.

    This is intended to work around issues where a date is not known for example but Python's date object requires
    one to be set. '1' is used as a default/convention but without a precision flag it isn't possible to know if '1'
    represents the first of the month (and is known), or represents an unknown value.

    This method is the inverse of `encode_date_string`.

    Examples
    --------
    * 'date(2012, 4, 18)' is returned as '2012-04-18'
    * 'datetime(2012, 4, 18, 22, 48, 56)' is returned as '2012-4-18T22:48:56'
    * 'date(2012, 4, 18), date_precision='year' is returned as '2012'
    * 'date(2012, 4, 18), date_precision='month' is returned as '2012-04'

    :param date_datetime: python date/datetime
    :param date_precision: qualifier to limit the precision of the date_datetime to a month or year
    :return: ISO 8601 formatted date/datetime

    """
    if date_precision == "year":
        return str(date_datetime.year)
    if date_precision == "month":
        return f"{date_datetime.year}-{date_datetime.month:02}"

    return date_datetime.isoformat()


def decode_date_string(date_datetime: str) -> dict:
    """
    Parses an ISO 8601 date, partial date or datetime string representation as a python date or datetime object.

    This method includes support for partial dates (year or year month) via an optional date precision element.
    When applicable, a `date_precision` property will be included in the returned dict to indicate the date object
    is precise to the 'month' or 'year'.

    This is intended to work around issues where a date is not known for example but Python's date object requires
    one to be set. '1' is used as a default/convention but without a precision flag it isn't possible to know if '1'
    represents the first of the month (and is known), or represents an unknown value.

    This method is the inverse of `encode_date_string`.

    Examples
    --------
    * '2012-04-18' is returned as '{date(2012, 4, 18)}'
    * '2012-4-18T22:48:56' is returned as 'datetime(2012, 4, 18, 22, 48, 56)'
    * '2012' is returned as {'date': date(2012, 1, 1), date_precision='year'}
    * '2012-04' is returned as {'date': date(2012, 1, 1), date_precision='month'}

    :param date_datetime: ISO 8601 formatted date/datetime
    :return: dict containing a python date/datetime and optionally a date_precision qualifying string

    """
    if "T" in date_datetime:
        return {"date": datetime.fromisoformat(date_datetime)}

    _ = {}
    _date_datetime_parts = date_datetime.split("-")
    if len(_date_datetime_parts) == 1:
        # Assume a year only date
        date_datetime = f"{date_datetime}-01-01"
        _["date_precision"] = "year"
    elif len(_date_datetime_parts) == 2:
        # Assume a year and month only date
        date_datetime = f"{date_datetime}-01"
        _["date_precision"] = "month"

    _["date"] = datetime.fromisoformat(date_datetime).date()
    return _


def condense_contacts_roles(contacts: list[dict]) -> list[dict]:
    """
    Groups separate contacts with multiple roles into a single contact with multiple roles.

    I.e. If two contacts are identical but with different, singular, roles, this method will return a single contact
    with multiple roles.

    E.g. A set of contacts: {'name': 'foo', role: ['a']}, {'name': 'foo', role: ['b']}, {'name': 'bar', role: ['a']}
               will become: {'name': 'foo', role: ['a', 'b']}, {'name': 'bar', role: ['a']}

    Note: this method triggers a bug-bear error in flake8 for an unused loop control variable. In this case the error
    is invalid as the 'key' control variable is used in the groupby lambda function, rather than the body of the loop.

    :param contacts: list of contacts to be grouped/reduced
    :return list of contacts with merged roles
    """
    _merged_contacts = []
    _contacts_without_roles = []
    for contact in contacts:
        _contact = deepcopy(contact)
        del _contact["role"]
        _contacts_without_roles.append(
            {"key": json.dumps(_contact), "key_data": _contact, "role_data": contact["role"][0]}
        )

    for key, contact in groupby(_contacts_without_roles, key=lambda x: x["key"]):  # noqa: B007 - see docblock
        contact = list(contact)
        _merged_contact = contact[0]["key_data"]
        _merged_contact["role"] = []
        for _contact in contact:
            _merged_contact["role"].append(_contact["role_data"])
        _merged_contacts.append(_merged_contact)

    return _merged_contacts


def condense_distribution_distributors(distributions: list[dict]) -> list[dict]:
    """
    Groups distribution options based on their distributor.

    Needed to encode distribution information in an ISO compatible structure.

    Inverse of the `flatten_distribution_distributors()` method.

    I.e. If two distribution options have the same distributor, this method will return a set of distributions with a
    single distributor containing two distribution options.

    In abstract terms it translates from this structure:

        distribution:
            |- distribution option [distributor, transfer_option, ...]


    ... to this structure:

        distribution:
            |- distributor:
                |- distribution option [transfer_option, ... (!distributor)]

    E.g. A set of (simplified) distribution options:
        - {'distributor': 'foo', 'transfer_option': {'href': 'https://example.com/a'}}
        - {'distributor': 'foo', 'transfer_option': {'href': 'https://example.com/b'}}
        - {'distributor': 'bar', 'transfer_option': {'href': 'https://example.com/c'}}
    ...will become:
        - {'distributor': 'foo', 'distribution_options': [
            {'transfer_option': {'href': 'https://example.com/a'}},
            {'transfer_option': {'href': 'https://example.com/b'}}
          ]}
        - {'distributor': 'bar', 'distribution_options': [{'transfer_option': {'href': 'https://example.com/c'}}]

    To clarify some terms used above:

    - distribution: the ISO `gmd:distributionInfo` element, which can contain a number of distribution elements
    - distributor, transfer_option, format: ISO distribution elements
    - distribution options: a non-ISO concept for a grouping of related distribution elements (can be incomplete)

    Note: distributors need to match exactly (i.e. '==' not 'like' or 'contains')

    E.g. a structure like this:
        - {'distributor': {'name: 'foo', 'address': '...'}, 'transfer_option': {'href': 'https://example.com/a'}}
        - {'distributor': {'name: 'foo'}, 'transfer_option': {'href': 'https://example.com/b'}}
    ... will become:
        - {'distributor': {'name: 'foo', 'address': '...'}, 'distribution_options': [
            {'transfer_option': {'href': 'https://example.com/a'}}
          ]}
        - {'distributor': {'name: 'foo'}, 'distributor': 'distribution_options': [
            {'transfer_option': {'href': 'https://example.com/b'}}
          ]}
    ... not:
        - {'distributor': {'name: 'foo', 'address': '...'}, 'distribution_options': [
            {'transfer_option': {'href': 'https://example.com/a'}},
            {'transfer_option': {'href': 'https://example.com/b'}}
          ]}

    :param distributions: list of distributions (distribution options)
    :return list of distributions (distribution options) grouped by distributor
    """
    _merged_distributor_distributions: dict[str, dict[str, list[dict]]] = {}

    for distribution_option in distributions:
        _distribution_option = deepcopy(distribution_option)
        del _distribution_option["distributor"]
        _distributor_key = json.dumps(distribution_option["distributor"])

        if _distributor_key not in _merged_distributor_distributions:
            _merged_distributor_distributions[_distributor_key] = {
                "distributor": distribution_option["distributor"],
                "distribution_options": [],
            }
        _merged_distributor_distributions[_distributor_key]["distribution_options"].append(_distribution_option)

    return list(_merged_distributor_distributions.values())


def flatten_distribution_distributors(distributions: list[dict]) -> list[dict]:
    """
    Flattens distribution options that are grouped by their distributor.

    In order to make it easy to update specific distribution elements in configs (by avoiding nesting).

    Inverse of the `condense_distribution_distributors()` method.

    In abstract terms this method translates from this structure:

        distribution:
            |- distributor:
                |- distribution option [transfer_option, ...]

    ... to this structure:
        distribution:
            |- distribution option [distributor, transfer_option, ...]

    See the `condense_distribution_distributors()` method for more examples (which will have the reverse effect).

    :param distributions: list of distributions (distribution options) grouped by distributor
    :return list of distributions (distribution options)
    """
    _flattened_distribution_options: list[dict] = []

    for distribution in distributions:
        for distribution_option in distribution["distribution_options"]:
            _distribution_option = deepcopy(distribution_option)
            _distribution_option["distributor"] = distribution["distributor"]
            _flattened_distribution_options.append(_distribution_option)

    return _flattened_distribution_options


def format_numbers_consistently(number: int | float) -> int | float:
    """
    Formats numeric values in a consistent way.

    Prevents inconsistencies with how numbers are formatted (e.g. '12.0' as '12')

    :param number: numeric value to format
    :return number as an integer if applicable, otherwise float
    """
    number = float(number)
    if number.is_integer():
        number = int(number)
    return number


def format_distribution_option_consistently(distribution_option: dict) -> dict:
    """
    Formats a distribution option object into a consistent structure.

    Distribution option objects are hashed to generate ID values for linking their format and transfer options together.
    As these hashes are sensitive to the order of information, and the format of numeric values etc., this method is
    used to sort and format these objects, so they are consistent, and therefore generate the same hash values.

    For example these serialised (simplified) objects are the same but have different hash values:

    1. '{'distributor': 'foo', 'format': 'csv', 'transfer_option': {'size': '40.0', 'url': 'https://example.com/foo.csv'}}'
        -> SHA1: e1342bc7fd5736aaf2cf7e6fd465c71d975b9747
    2. '{'transfer_option': {'size': '40', 'url': 'https://example.com/foo.csv'}, 'distributor': 'foo', 'format': 'csv'}}' ->
        SHA1: e4554470d7712aadf5f5467d6bd1427a689dea74

    By sorting the keys (so 'format' always comes before 'transfer_option' for example) and formatting '40.0' as '40',
    these differences are removed and the same hash value is given.

    :param distribution_option: distribution option object
    :return consistently structured/formatted distribution option object
    """
    _distribution_option = deepcopy(distribution_option)
    _distribution_option = _sort_dict_by_keys(dictionary=_distribution_option)
    if (
        "transfer_option" in _distribution_option
        and "size" in _distribution_option["transfer_option"]
        and "magnitude" in _distribution_option["transfer_option"]["size"]
    ):
        _distribution_option["transfer_option"]["size"]["magnitude"] = format_numbers_consistently(
            number=_distribution_option["transfer_option"]["size"]["magnitude"]
        )

    return _distribution_option


def decode_config_from_json(config: dict) -> dict:
    """
    Parse a record configuration loaded from a JSON encoded document.

    Specifically this method looks for any string encoded date or datetime values and converts them to their Python
    equivalents, including workarounds for partial dates if applicable. E.g. '2012-02' becomes
    `{'date': date(2012, 2, 1), 'date_precision': 'year'}`.

    This method is the reverse of `encode_config_for_json()`.

    :param config: record configuration
    :return parsed record configuration
    """
    return _decode_date_properties(dictionary=config)


def encode_config_for_json(config: dict) -> dict:
    """
    Prepare a record configuration for use in a JSON encoded document.

    Specifically this method looks for any date or datetime values and converts them to their string equivalents.
    E.g. `{'date': date(2012, 2, 1), 'date_precision': 'year'}` becomes '2012-02'.

    This method is the reverse of `decode_config_from_json()`.

    :param config: record configuration
    :return encoded record configuration
    """
    return _encode_date_properties(dictionary=config)


iso_19115_0_v3 = (
    "https://metadata-standards.data.bas.ac.uk/bas-metadata-generator-configuration-schemas/v2/iso-19115-1-v3.json"
)
iso_19115_2_v3 = (
    "https://metadata-standards.data.bas.ac.uk/bas-metadata-generator-configuration-schemas/v2/iso-19115-2-v3.json"
)
iso_19115_0_v4 = (
    "https://metadata-standards.data.bas.ac.uk/bas-metadata-generator-configuration-schemas/v2/iso-19115-0-v4.json"
)
iso_19115_2_v4 = (
    "https://metadata-standards.data.bas.ac.uk/bas-metadata-generator-configuration-schemas/v2/iso-19115-2-v4.json"
)



from __future__ import annotations

import json
from pathlib import Path
from typing import Any

from importlib_resources import files as resource_file
from lxml.etree import Element, SubElement, fromstring

from bas_metadata_library import (
    MetadataRecord as _MetadataRecord,
)
from bas_metadata_library import (
    MetadataRecordConfig as _MetadataRecordConfig,
)
from bas_metadata_library import (
    MetadataRecordElement as _MetadataRecordElement,
)
from bas_metadata_library import (
    Namespaces as _Namespaces,
)
from bas_metadata_library.standards.iec_pas_61174_common.utils import (
    generate_rtzp_archive as _generate_rtzp_archive,
)
from bas_metadata_library.standards.iec_pas_61174_common.utils import (
    load_record_from_rtzp_archive,
)


class Namespaces(_Namespaces):
    """Defines the namespaces for this standard."""

    rtz = "http://www.cirm.org/RTZ/1/2"
    xsi = "http://www.w3.org/2001/XMLSchema-instance"

    _schema_locations = {  # noqa: RUF012
        "rtz": "https://www.cirm.org/rtz/RTZ%20Schema%20version%201_2.xsd",
    }

    def __init__(self):
        self._namespaces = {
            "rtz": self.rtz,
            "xsi": self.xsi,
        }

        super().__init__(namespaces=self._namespaces, root_namespace=self.rtz)


class MetadataRecordElement(_MetadataRecordElement):
    """
    Overloaded base MetadataRecordElement class.

    Sets:
     - the type hint of the record attribute to the MetadataRecord class for this metadata standard
     - the namespaces property to the Namespaces class for this metadata standard
    """

    def __init__(
        self,
        record: _MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=xpath,
        )
        self.ns = Namespaces()


class MetadataRecordConfigV1(_MetadataRecordConfig):
    """
    Overloaded base MetadataRecordConfig class.

    Defines version 1 of the JSON Schema used for this metadata standard.
    """

    def __init__(self, **kwargs: dict):
        super().__init__(**kwargs)

        self.config = kwargs

        schema_path = resource_file("bas_metadata_library.schemas.dist").joinpath("iec_pas_61174_1_v1.json")
        with schema_path.open() as schema_file:
            schema_data = json.load(schema_file)
        self.schema = schema_data

        # Workaround - will be addressed in #149
        self.schema_uri = schema_data["$id"]
        self.config = {"$schema": self.schema_uri, **kwargs}


class MetadataRecord(_MetadataRecord):
    """Defines the root element, and it's sub-elements, for this metadata standard."""

    def __init__(self, configuration: MetadataRecordConfigV1 = None, record: str | None = None):
        self.ns = Namespaces()
        self.attributes = {}
        self.record = Element(
            f"{{{self.ns.rtz}}}route",
            attrib={
                "version": str(1.2),
                f"{{{self.ns.xsi}}}schemaLocation": self.ns.schema_locations(),
            },
            nsmap=self.ns.nsmap(),
        )
        self.xpath = "/rtz:route"

        if configuration is not None:
            configuration.validate()
            self.attributes: dict[str, Any] = configuration.config

        if record is not None:
            self.record = fromstring(record.encode())  # noqa: S320 (see '`lxml` package (security)' README section)

        self.metadata_record = Route(record=self.record, attributes=self.attributes, xpath=self.xpath)

    def make_config(self) -> MetadataRecordConfigV1:
        return MetadataRecordConfigV1(**self.metadata_record.make_config())

    def make_element(self) -> Element:
        return self.metadata_record.make_element()

    def generate_rtzp_archive(self, file: Path) -> None:
        """
        Generates RTZ/XML record in a RTZP data container.

        A RTZP container is a zip archive containing a single RTZ file.

        :param file: path at which to create RTZP data container
        """
        _generate_rtzp_archive(
            file=file, rtz_name=self.attributes["route_name"], rtz_document=self.generate_xml_document().decode()
        )

    def load_from_rtzp_archive(self, file: Path) -> None:
        """Loads a RTZ record from a RTZP data container."""
        self.record = load_record_from_rtzp_archive(file=file)
        self.metadata_record = Route(record=self.record, attributes=self.attributes, xpath=self.xpath)

    # noinspection PyMethodOverriding
    def validate(self) -> None:
        super().validate(xsd_path=Path("rtz/rtz_1_2.xsd"))


class Route(MetadataRecordElement):
    def make_config(self) -> dict:
        _ = {}

        route_information = RouteInformation(record=self.record, attributes=self.attributes, xpath=f"{self.xpath}")
        _route_information = route_information.make_config()
        if bool(_route_information):
            if "author" in _route_information:
                _["route_author"] = _route_information["author"]
            if "name" in _route_information:
                _["route_name"] = _route_information["name"]
            if "status" in _route_information:
                _["route_status"] = _route_information["status"]

        waypoints = Waypoints(record=self.record, attributes=self.attributes, xpath=f"{self.xpath}")
        _waypoints = waypoints.make_config()
        if bool(_waypoints):
            _["waypoints"] = _waypoints

        return _

    def make_element(self) -> None:
        if (
            "route_author" in self.element_attributes
            or "route_name" in self.element_attributes
            or "route_status" in self.element_attributes
        ):
            route_information = RouteInformation(
                record=self.record,
                attributes=self.attributes,
                element_attributes=self.attributes,
                parent_element=self.record,
            )
            route_information.make_element()
        if "waypoints" in self.element_attributes:
            waypoints = Waypoints(
                record=self.record,
                attributes=self.attributes,
                element_attributes=self.attributes,
                parent_element=self.record,
            )
            waypoints.make_element()

        return self.record


class RouteInformation(MetadataRecordElement):
    def make_config(self) -> dict:
        _ = {}

        route_author = self.record.xpath(
            f"{self.xpath}/rtz:routeInfo/@routeAuthor",
            namespaces=self.ns.nsmap(suppress_root_namespace=True),
        )
        if len(route_author) > 0:
            _["author"] = route_author[0]

        route_name = self.record.xpath(
            f"{self.xpath}/rtz:routeInfo/@routeName",
            namespaces=self.ns.nsmap(suppress_root_namespace=True),
        )
        if len(route_name) > 0:
            _["name"] = route_name[0]

        route_status = self.record.xpath(
            f"{self.xpath}/rtz:routeInfo/@routeStatus",
            namespaces=self.ns.nsmap(suppress_root_namespace=True),
        )
        if len(route_status) > 0:
            _["status"] = route_status[0]

        return _

    def make_element(self) -> None:
        attributes = {}

        if "route_author" in self.element_attributes:
            attributes["routeAuthor"] = self.element_attributes["route_author"]
        if "route_name" in self.element_attributes:
            attributes["routeName"] = self.element_attributes["route_name"]
        if "route_status" in self.element_attributes:
            attributes["routeStatus"] = self.element_attributes["route_status"]

        SubElement(self.parent_element, f"{{{self.ns.rtz}}}routeInfo", attrib=attributes)


class Waypoints(MetadataRecordElement):
    def make_config(self) -> list:
        _ = []

        waypoints_length = int(
            self.record.xpath(
                f"count({self.xpath}/rtz:waypoints/rtz:waypoint)",
                namespaces=self.ns.nsmap(suppress_root_namespace=True),
            )
        )
        for waypoint_index in range(1, waypoints_length + 1):
            _waypoint = {}

            waypoint_id = self.record.xpath(
                f"{self.xpath}/rtz:waypoints/rtz:waypoint[{waypoint_index}]/@id",
                namespaces=self.ns.nsmap(suppress_root_namespace=True),
            )
            if len(waypoint_id) > 0:
                _waypoint["id"] = int(waypoint_id[0])

            waypoint_revision = self.record.xpath(
                f"{self.xpath}/rtz:waypoints/rtz:waypoint[{waypoint_index}]/@revision",
                namespaces=self.ns.nsmap(suppress_root_namespace=True),
            )
            if len(waypoint_revision) > 0:
                _waypoint["revision"] = int(waypoint_revision[0])

            position = Position(
                record=self.record,
                attributes=self.attributes,
                xpath=f"{self.xpath}/rtz:waypoints/rtz:waypoint[{waypoint_index}]",
            )
            _position = position.make_config()
            if bool(_position):
                _waypoint["position"] = _position

            leg = self.record.xpath(
                f"{self.xpath}/rtz:waypoints/rtz:waypoint[{waypoint_index}]/rtz:leg",
                namespaces=self.ns.nsmap(suppress_root_namespace=True),
            )
            if len(leg) > 0:
                leg = Leg(
                    record=self.record,
                    attributes=self.attributes,
                    xpath=f"{self.xpath}/rtz:waypoints/rtz:waypoint[{waypoint_index}]",
                )
                _leg = leg.make_config()
                if bool(_leg):
                    _waypoint["leg"] = _leg

            if bool(_waypoint):
                _.append(_waypoint)

        return _

    def make_element(self) -> None:
        waypoints_element = SubElement(self.parent_element, f"{{{self.ns.rtz}}}waypoints")
        for waypoint in self.element_attributes["waypoints"]:
            attributes = {"id": str(waypoint["id"]), "revision": str(waypoint["revision"])}
            waypoint_element = SubElement(waypoints_element, f"{{{self.ns.rtz}}}waypoint", attrib=attributes)

            position = Position(
                record=self.record,
                attributes=self.attributes,
                element_attributes=waypoint["position"],
                parent_element=waypoint_element,
            )
            position.make_element()

            if "leg" in waypoint:
                leg = Leg(
                    record=self.record,
                    attributes=self.attributes,
                    element_attributes=waypoint["leg"],
                    parent_element=waypoint_element,
                )
                leg.make_element()


class Position(MetadataRecordElement):
    def make_config(self) -> dict:
        _ = {}

        lat = self.record.xpath(
            f"{self.xpath}/rtz:position/@lat",
            namespaces=self.ns.nsmap(suppress_root_namespace=True),
        )
        if len(lat) > 0:
            _["lat"] = float(lat[0])

        lon = self.record.xpath(
            f"{self.xpath}/rtz:position/@lon",
            namespaces=self.ns.nsmap(suppress_root_namespace=True),
        )
        if len(lon) > 0:
            _["lon"] = float(lon[0])

        return _

    def make_element(self) -> None:
        attributes = {"lat": str(self.element_attributes["lat"]), "lon": str(self.element_attributes["lon"])}
        SubElement(self.parent_element, f"{{{self.ns.rtz}}}position", attrib=attributes)


class Leg(MetadataRecordElement):
    def make_config(self) -> dict:
        _ = {}

        geometry_type = self.record.xpath(
            f"{self.xpath}/rtz:leg/@geometryType",
            namespaces=self.ns.nsmap(suppress_root_namespace=True),
        )
        if len(geometry_type) > 0:
            _["geometry_type"] = geometry_type[0]

        return _

    def make_element(self) -> None:
        attributes = {}
        if "geometry_type" in self.element_attributes:
            attributes["geometryType"] = self.element_attributes["geometry_type"]
        SubElement(self.parent_element, f"{{{self.ns.rtz}}}leg", attrib=attributes)



from bas_metadata_library.standards.iso_19115_common import MetadataRecordElement
from bas_metadata_library.standards.iso_19115_common.base_elements import (
    Contact,
    DateStamp,
    FileIdentifier,
    HierarchyLevel,
    MetadataMaintenance,
    MetadataStandard,
    ReferenceSystemInfo,
)
from bas_metadata_library.standards.iso_19115_common.common_elements import CharacterSet, Language
from bas_metadata_library.standards.iso_19115_common.data_distribution_elements import DataDistribution
from bas_metadata_library.standards.iso_19115_common.data_identification_elements import DataIdentification
from bas_metadata_library.standards.iso_19115_common.data_quality_elements import DataQuality
from bas_metadata_library.standards.iso_19115_common.utils import condense_contacts_roles


class ISOMetadataRecord(MetadataRecordElement):
    def make_config(  # noqa: C901 see uk-pdc/metadata-infrastructure/metadata-library#175 for more information
        self,
    ) -> dict:
        _ = {}

        file_identifier = FileIdentifier(record=self.record, attributes=self.attributes, xpath=f"{self.xpath}")
        _file_identifier = file_identifier.make_config()
        if _file_identifier != "":
            # noinspection PyTypeChecker
            _["file_identifier"] = _file_identifier

        language = Language(record=self.record, attributes=self.attributes, xpath=f"{self.xpath}/gmd:language")
        _language = language.make_config()
        if _language != "":
            if "metadata" not in _:
                _["metadata"] = {}
            _["metadata"]["language"] = _language

        character_set = CharacterSet(
            record=self.record, attributes=self.attributes, xpath=f"{self.xpath}/gmd:characterSet"
        )
        _character_set = character_set.make_config()
        if _character_set != "":
            if "metadata" not in _:
                _["metadata"] = {}
            _["metadata"]["character_set"] = _character_set

        hierarchy_level = HierarchyLevel(
            record=self.record, attributes=self.attributes, xpath=f"/{self.xpath}/gmd:hierarchyLevel"
        )
        _hierarchy_level = hierarchy_level.make_config()
        if _hierarchy_level != "":
            if "metadata" not in _:
                _["metadata"] = {}
            _["hierarchy_level"] = _hierarchy_level

        _contacts = []
        contacts_length = int(
            self.record.xpath(
                f"count({self.xpath}/gmd:contact)",
                namespaces=self.ns.nsmap(),
            )
        )
        for contact_index in range(1, contacts_length + 1):
            contact = Contact(
                record=self.record, attributes=self.attributes, xpath=f"({self.xpath}/gmd:contact)[{contact_index}]"
            )
            _contact = contact.make_config
            if bool(_contact):
                _contacts.append(_contact)
        if len(_contacts) > 0:
            if "metadata" not in _:
                _["metadata"] = {}
            _["metadata"]["contacts"] = _contacts

        date_stamp = DateStamp(record=self.record, attributes=self.attributes, xpath=f"{self.xpath}")
        _date_stamp = date_stamp.make_config()
        if _date_stamp is not None:
            _["metadata"]["date_stamp"] = _date_stamp

        metadata_standard = MetadataStandard(record=self.record, attributes=self.attributes, xpath=f"{self.xpath}")
        _metadata_standard = metadata_standard.make_config()
        if bool(_metadata_standard):
            _["metadata"]["metadata_standard"] = _metadata_standard

        reference_system_identifier = ReferenceSystemInfo(
            record=self.record, attributes=self.attributes, xpath=f"{self.xpath}"
        )
        _reference_system_identifier = reference_system_identifier.make_config()
        if bool(_reference_system_identifier):
            # noinspection PyTypeChecker
            _["reference_system_info"] = _reference_system_identifier

        data_identification = DataIdentification(record=self.record, attributes=self.attributes, xpath=f"{self.xpath}")
        _data_identification = data_identification.make_config()
        if bool(_data_identification):
            if "identification" not in _:
                _["identification"] = {}
            _["identification"] = {**_["identification"], **_data_identification}

        data_distribution = DataDistribution(record=self.record, attributes=self.attributes, xpath=f"{self.xpath}")
        _data_distribution = data_distribution.make_config()
        if len(_data_distribution) > 0:
            _["distribution"] = _data_distribution

        data_quality = DataQuality(record=self.record, attributes=self.attributes, xpath=f"{self.xpath}")
        _data_quality = data_quality.make_config()
        if bool(_data_quality):
            _["identification"] = {**_["identification"], **_data_quality}

        metadata_maintenance = MetadataMaintenance(
            record=self.record, attributes=self.attributes, xpath=f"{self.xpath}"
        )
        _metadata_maintenance = metadata_maintenance.make_config()
        if bool(_metadata_maintenance):
            _["metadata"]["maintenance"] = _metadata_maintenance

        if "identification" in _ and "contacts" in _["identification"]:
            _["identification"]["contacts"] = condense_contacts_roles(contacts=_["identification"]["contacts"])

        return _

    def make_element(self) -> None:  # noqa: C901 see uk-pdc/metadata-infrastructure/metadata-library#175
        if "file_identifier" in self.attributes:
            identifier = FileIdentifier(record=self.record, attributes=self.attributes, parent_element=self.record)
            identifier.make_element()

        if "metadata" in self.attributes and "language" in self.attributes["metadata"]:
            language = Language(record=self.record, attributes=self.attributes["metadata"])
            language.make_element()

        if "metadata" in self.attributes and "character_set" in self.attributes["metadata"]:
            character_set = CharacterSet(
                record=self.record, attributes=self.attributes["metadata"], xpath=f"{self.xpath}"
            )
            character_set.make_element()

        if "hierarchy_level" in self.attributes:
            hierarchy_level = HierarchyLevel(record=self.record, attributes=self.attributes)
            hierarchy_level.make_element()

        if "metadata" in self.attributes and "contacts" in self.attributes["metadata"]:
            for contact_attributes in self.attributes["metadata"]["contacts"]:
                for role in contact_attributes["role"]:
                    _contact = contact_attributes.copy()
                    _contact["role"] = role

                    contact = Contact(
                        record=self.record,
                        attributes=self.attributes,
                        parent_element=self.record,
                        element_attributes=_contact,
                    )
                    contact.make_element()

        if "metadata" in self.attributes and "date_stamp" in self.attributes["metadata"]:
            date_stamp = DateStamp(record=self.record, attributes=self.attributes["metadata"])
            date_stamp.make_element()

        if "metadata" in self.attributes and "metadata_standard" in self.attributes["metadata"]:
            metadata_standard = MetadataStandard(
                record=self.record,
                attributes=self.attributes,
                parent_element=self.record,
                element_attributes=self.attributes["metadata"]["metadata_standard"],
            )
            metadata_standard.make_element()

        if "reference_system_info" in self.attributes:
            reference_system_info = ReferenceSystemInfo(
                record=self.record,
                attributes=self.attributes,
                parent_element=self.record,
                element_attributes=self.attributes["reference_system_info"],
            )
            reference_system_info.make_element()

        if "identification" in self.attributes:
            data_identification = DataIdentification(record=self.record, attributes=self.attributes)
            data_identification.make_element()

        if "distribution" in self.attributes:
            data_distribution = DataDistribution(
                record=self.record, attributes=self.attributes, element_attributes=self.attributes["distribution"]
            )
            data_distribution.make_element()

        if ("hierarchy_level" in self.attributes) or (
            "identification" in self.attributes
            and ("measures" in self.attributes["identification"] or "lineage" in self.attributes["identification"])
        ):
            data_quality = DataQuality(record=self.record, attributes=self.attributes)
            data_quality.make_element()

        if "metadata" in self.attributes and "maintenance" in self.attributes["metadata"]:
            metadata_maintenance = MetadataMaintenance(
                record=self.record,
                attributes=self.attributes,
                parent_element=self.record,
                element_attributes=self.attributes["metadata"]["maintenance"],
            )
            metadata_maintenance.make_element()

        return self.record



from __future__ import annotations

from copy import deepcopy
from datetime import datetime

from lxml.etree import Element, SubElement

from bas_metadata_library import MetadataRecord
from bas_metadata_library import MetadataRecord as _MetadataRecord
from bas_metadata_library.standards.iso_19115_common import CodeListElement, MetadataRecordElement
from bas_metadata_library.standards.iso_19115_common.utils import decode_date_string, encode_date_string


class Language(CodeListElement):
    def __init__(
        self,
        record: MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=f"{xpath}/gmd:LanguageCode",
        )
        self.code_list_values = ["eng"]
        self.code_list = "http://www.loc.gov/standards/iso639-2/php/code_list.php"
        self.element = f"{{{self.ns.gmd}}}language"
        self.element_code = f"{{{self.ns.gmd}}}LanguageCode"
        self.attribute = "language"


class CharacterSet(CodeListElement):
    def __init__(
        self,
        record: MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=f"{xpath}/gmd:MD_CharacterSetCode",
        )
        self.code_list_values = ["utf8"]
        self.code_list = (
            "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/"
            "codelist/gmxCodelists.xml#MD_CharacterSetCode"
        )
        self.element = f"{{{self.ns.gmd}}}characterSet"
        self.element_code = f"{{{self.ns.gmd}}}MD_CharacterSetCode"
        self.attribute = "character_set"


class ResponsibleParty(MetadataRecordElement):
    def __init__(
        self,
        record: _MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(record, attributes, parent_element, element_attributes, xpath)

    def make_config(  # noqa: C901 see uk-pdc/metadata-infrastructure/metadata-library#175 for more information
        self,
    ) -> dict:
        _ = {}

        individual_name = self.record.xpath(
            f"{ self.xpath }/gmd:CI_ResponsibleParty/gmd:individualName/gmx:Anchor/text() | "
            f"{ self.xpath }/gmd:CI_ResponsibleParty/gmd:individualName/gco:CharacterString/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(individual_name) > 0:
            if "individual" not in _:
                _["individual"] = {}
            _["individual"]["name"] = individual_name[0]

        individual_href = self.record.xpath(
            f"{ self.xpath }/gmd:CI_ResponsibleParty/gmd:individualName/gmx:Anchor/@xlink:href",
            namespaces=self.ns.nsmap(),
        )
        if len(individual_href) > 0:
            if "individual" not in _:
                _["individual"] = {}
            _["individual"]["href"] = individual_href[0]

        individual_title = self.record.xpath(
            f"{ self.xpath }/gmd:CI_ResponsibleParty/gmd:individualName/gmx:Anchor/@xlink:title",
            namespaces=self.ns.nsmap(),
        )
        if len(individual_title) > 0:
            _["individual"]["title"] = individual_title[0]

        organisation_name = self.record.xpath(
            f"{ self.xpath }/gmd:CI_ResponsibleParty/gmd:organisationName/gmx:Anchor/text() | "
            f"{ self.xpath }/gmd:CI_ResponsibleParty/gmd:organisationName/gco:CharacterString/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(organisation_name) > 0:
            if "organisation" not in _:
                _["organisation"] = {}
            _["organisation"]["name"] = organisation_name[0]

        organisation_href = self.record.xpath(
            f"{ self.xpath }/gmd:CI_ResponsibleParty/gmd:organisationName/gmx:Anchor/@xlink:href",
            namespaces=self.ns.nsmap(),
        )
        if len(organisation_href) > 0:
            if "organisation" not in _:
                _["organisation"] = {}
            _["organisation"]["href"] = organisation_href[0]

        organisation_title = self.record.xpath(
            f"{ self.xpath }/gmd:CI_ResponsibleParty/gmd:organisationName/gmx:Anchor/@xlink:title",
            namespaces=self.ns.nsmap(),
        )
        if len(organisation_title) > 0:
            _["organisation"]["title"] = organisation_title[0]

        position_value = self.record.xpath(
            f"{self.xpath}/gmd:CI_ResponsibleParty/gmd:positionName/gco:CharacterString/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(position_value) > 0:
            _["position"] = position_value[0]

        phone_value = self.record.xpath(
            f"{self.xpath}/gmd:CI_ResponsibleParty/gmd:contactInfo/gmd:CI_Contact/gmd:phone/"
            f"gmd:CI_Telephone/gmd:voice/gco:CharacterString/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(phone_value) > 0:
            _["phone"] = phone_value[0]

        delivery_point_value = self.record.xpath(
            f"{self.xpath}/gmd:CI_ResponsibleParty/gmd:contactInfo/gmd:CI_Contact/gmd:address/"
            f"gmd:CI_Address/gmd:deliveryPoint/gco:CharacterString/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(delivery_point_value) > 0:
            if "address" not in _:
                _["address"] = {}
            _["address"]["delivery_point"] = delivery_point_value[0]

        city_value = self.record.xpath(
            f"{self.xpath}/gmd:CI_ResponsibleParty/gmd:contactInfo/gmd:CI_Contact/gmd:address/"
            f"gmd:CI_Address/gmd:city/gco:CharacterString/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(city_value) > 0:
            if "address" not in _:
                _["address"] = {}
            _["address"]["city"] = city_value[0]

        administrative_area_value = self.record.xpath(
            f"{self.xpath}/gmd:CI_ResponsibleParty/gmd:contactInfo/gmd:CI_Contact/gmd:address/"
            f"gmd:CI_Address/gmd:administrativeArea/gco:CharacterString/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(administrative_area_value) > 0:
            if "address" not in _:
                _["address"] = {}
            _["address"]["administrative_area"] = administrative_area_value[0]

        postal_code_value = self.record.xpath(
            f"{self.xpath}/gmd:CI_ResponsibleParty/gmd:contactInfo/gmd:CI_Contact/gmd:address/"
            f"gmd:CI_Address/gmd:postalCode/gco:CharacterString/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(postal_code_value) > 0:
            if "address" not in _:
                _["address"] = {}
            _["address"]["postal_code"] = postal_code_value[0]

        country_value = self.record.xpath(
            f"{self.xpath}/gmd:CI_ResponsibleParty/gmd:contactInfo/gmd:CI_Contact/gmd:address/"
            f"gmd:CI_Address/gmd:country/gco:CharacterString/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(country_value) > 0:
            if "address" not in _:
                _["address"] = {}
            _["address"]["country"] = country_value[0]

        email_value = self.record.xpath(
            f"{self.xpath}/gmd:CI_ResponsibleParty/gmd:contactInfo/gmd:CI_Contact/gmd:address/"
            f"gmd:CI_Address/gmd:electronicMailAddress/gco:CharacterString/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(email_value) > 0:
            _["email"] = email_value[0]

        online_resource = OnlineResource(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:CI_ResponsibleParty/gmd:contactInfo/gmd:CI_Contact/gmd:onlineResource",
        )
        _online_resource = online_resource.make_config()
        if bool(_online_resource):
            _["online_resource"] = _online_resource

        role = Role(
            record=self.record, attributes=self.attributes, xpath=f"{self.xpath}/gmd:CI_ResponsibleParty/gmd:role"
        )
        _role = role.make_config()
        if _role != "":
            _["role"] = [_role]

        return _

    def make_element(self) -> None:  # noqa: C901 see uk-pdc/metadata-infrastructure/metadata-library#175
        responsible_party_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}CI_ResponsibleParty")

        if "individual" in self.element_attributes and "name" in self.element_attributes["individual"]:
            individual_element = SubElement(responsible_party_element, f"{{{self.ns.gmd}}}individualName")
            if "href" in self.element_attributes["individual"]:
                anchor = AnchorElement(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=individual_element,
                    element_attributes=self.element_attributes["individual"],
                    element_value=self.element_attributes["individual"]["name"],
                )
                anchor.make_element()
            else:
                individual_value = SubElement(individual_element, f"{{{self.ns.gco}}}CharacterString")
                individual_value.text = self.element_attributes["individual"]["name"]

        if "organisation" in self.element_attributes and "name" in self.element_attributes["organisation"]:
            organisation_element = SubElement(responsible_party_element, f"{{{self.ns.gmd}}}organisationName")
            if "href" in self.element_attributes["organisation"]:
                anchor = AnchorElement(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=organisation_element,
                    element_attributes=self.element_attributes["organisation"],
                    element_value=self.element_attributes["organisation"]["name"],
                )
                anchor.make_element()
            else:
                organisation_name_value = SubElement(organisation_element, f"{{{self.ns.gco}}}CharacterString")
                organisation_name_value.text = self.element_attributes["organisation"]["name"]

        if "position" in self.element_attributes:
            position_name_element = SubElement(responsible_party_element, f"{{{self.ns.gmd}}}positionName")
            position_name_value = SubElement(position_name_element, f"{{{self.ns.gco}}}CharacterString")
            position_name_value.text = self.element_attributes["position"]

        if (
            "phone" in self.element_attributes
            or "address" in self.element_attributes
            or "email" in self.element_attributes
            or "online_resource" in self.element_attributes
        ):
            contact_wrapper = SubElement(responsible_party_element, f"{{{self.ns.gmd}}}contactInfo")
            contact_element = SubElement(contact_wrapper, f"{{{self.ns.gmd}}}CI_Contact")

            if "phone" in self.element_attributes:
                phone_wrapper = SubElement(contact_element, f"{{{self.ns.gmd}}}phone")
                phone_element = SubElement(phone_wrapper, f"{{{self.ns.gmd}}}CI_Telephone")
                phone_voice = SubElement(phone_element, f"{{{self.ns.gmd}}}voice")
                phone_voice_value = SubElement(phone_voice, f"{{{self.ns.gco}}}CharacterString")
                phone_voice_value.text = self.element_attributes["phone"]

            if "address" in self.element_attributes or "email" in self.element_attributes:
                address_wrapper = SubElement(contact_element, f"{{{self.ns.gmd}}}address")
                address_element = SubElement(address_wrapper, f"{{{self.ns.gmd}}}CI_Address")

                if "address" in self.element_attributes:
                    if "delivery_point" in self.element_attributes["address"]:
                        delivery_point_element = SubElement(address_element, f"{{{self.ns.gmd}}}deliveryPoint")
                        delivery_point_value = SubElement(delivery_point_element, f"{{{self.ns.gco}}}CharacterString")
                        delivery_point_value.text = self.element_attributes["address"]["delivery_point"]
                    if "city" in self.element_attributes["address"]:
                        city_element = SubElement(address_element, f"{{{self.ns.gmd}}}city")
                        city_value = SubElement(city_element, f"{{{self.ns.gco}}}CharacterString")
                        city_value.text = self.element_attributes["address"]["city"]
                    if "administrative_area" in self.element_attributes["address"]:
                        administrative_area_element = SubElement(
                            address_element, f"{{{self.ns.gmd}}}administrativeArea"
                        )
                        administrative_area_value = SubElement(
                            administrative_area_element, f"{{{self.ns.gco}}}CharacterString"
                        )
                        administrative_area_value.text = self.element_attributes["address"]["administrative_area"]
                    if "postal_code" in self.element_attributes["address"]:
                        postal_code_element = SubElement(address_element, f"{{{self.ns.gmd}}}postalCode")
                        postal_code_value = SubElement(postal_code_element, f"{{{self.ns.gco}}}CharacterString")
                        postal_code_value.text = self.element_attributes["address"]["postal_code"]
                    if "country" in self.element_attributes["address"]:
                        country_element = SubElement(address_element, f"{{{self.ns.gmd}}}country")
                        country_value = SubElement(country_element, f"{{{self.ns.gco}}}CharacterString")
                        country_value.text = self.element_attributes["address"]["country"]

                if "email" in self.element_attributes:
                    email_element = SubElement(address_element, f"{{{self.ns.gmd}}}electronicMailAddress")
                    email_value = SubElement(email_element, f"{{{self.ns.gco}}}CharacterString")
                    email_value.text = self.element_attributes["email"]
                else:
                    SubElement(
                        address_element,
                        f"{{{self.ns.gmd}}}electronicMailAddress",
                        attrib={f"{{{self.ns.gco}}}nilReason": "unknown"},
                    )

            if "online_resource" in self.element_attributes:
                online_resource_wrapper = SubElement(contact_element, f"{{{self.ns.gmd}}}onlineResource")
                online_resource = OnlineResource(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=online_resource_wrapper,
                    element_attributes=self.element_attributes["online_resource"],
                )
                online_resource.make_element()

        if "role" in self.element_attributes:
            role = Role(
                record=self.record,
                attributes=self.attributes,
                parent_element=responsible_party_element,
                element_attributes=self.element_attributes,
            )
            role.make_element()


class OnlineResource(MetadataRecordElement):
    def make_config(self) -> dict:
        _ = {}

        linkage = Linkage(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{ self.xpath }/gmd:CI_OnlineResource/gmd:linkage",
        )
        _linkage = linkage.make_config()
        if "href" in _linkage:
            _["href"] = _linkage["href"]

        name_value = self.record.xpath(
            f"{self.xpath}/gmd:CI_OnlineResource/gmd:name/gco:CharacterString/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(name_value) > 0:
            _["title"] = name_value[0]

        description_value = self.record.xpath(
            f"{self.xpath}/gmd:CI_OnlineResource/gmd:description/gco:CharacterString/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(description_value) > 0:
            _["description"] = description_value[0]

        protocol_value = self.record.xpath(
            f"{self.xpath}/gmd:CI_OnlineResource/gmd:protocol/gco:CharacterString/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(protocol_value) > 0:
            _["protocol"] = protocol_value[0]

        function = OnlineRole(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:CI_OnlineResource/gmd:function",
        )
        _function = function.make_config()
        if _function != "":
            _["function"] = _function

        return _

    def make_element(self) -> None:
        online_resource_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}CI_OnlineResource")

        if "href" in self.element_attributes:
            linkage = Linkage(
                record=self.record,
                attributes=self.attributes,
                parent_element=online_resource_element,
                element_attributes=self.element_attributes,
            )
            linkage.make_element()

        if "protocol" in self.element_attributes:
            protocol_element = SubElement(online_resource_element, f"{{{self.ns.gmd}}}protocol")
            protocol_value = SubElement(protocol_element, f"{{{self.ns.gco}}}CharacterString")
            protocol_value.text = self.element_attributes["protocol"]

        if "title" in self.element_attributes:
            title_wrapper = SubElement(online_resource_element, f"{{{self.ns.gmd}}}name")
            title_element = SubElement(title_wrapper, f"{{{self.ns.gco}}}CharacterString")
            title_element.text = self.element_attributes["title"]

        if "description" in self.element_attributes:
            title_wrapper = SubElement(online_resource_element, f"{{{self.ns.gmd}}}description")
            title_element = SubElement(title_wrapper, f"{{{self.ns.gco}}}CharacterString")
            title_element.text = self.element_attributes["description"]

        if "function" in self.element_attributes:
            function = OnlineRole(
                record=self.record,
                attributes=self.attributes,
                parent_element=online_resource_element,
                element_attributes=self.element_attributes,
            )
            function.make_element()


class Linkage(MetadataRecordElement):
    def make_config(self) -> dict:
        _ = {}

        url_value = self.record.xpath(
            f"{self.xpath}/gmd:URL/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(url_value) > 0:
            _["href"] = url_value[0]

        return _

    def make_element(self) -> None:
        linkage_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}linkage")
        if "href" in self.element_attributes:
            url_value = SubElement(linkage_element, f"{{{self.ns.gmd}}}URL")
            url_value.text = self.element_attributes["href"]


class Role(CodeListElement):
    def __init__(
        self,
        record: MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=f"{xpath}/gmd:CI_RoleCode",
        )
        self.code_list_values = [
            "author",
            "custodian",
            "distributor",
            "originator",
            "owner",
            "pointOfContact",
            "principalInvestigator",
            "processor",
            "publisher",
            "resourceProvider",
            "sponsor",
            "user",
            "coAuthor",
            "collaborator",
            "contributor",
            "editor",
            "funder",
            "mediator",
            "rightsHolder",
            "stakeholder",
        ]
        self.code_list = "https://standards.iso.org/iso/19115/resources/Codelists/cat/codelists.xml#CI_RoleCode"
        self.element = f"{{{self.ns.gmd}}}role"
        self.element_code = f"{{{self.ns.gmd}}}CI_RoleCode"
        self.attribute = "role"


class OnlineRole(CodeListElement):
    def __init__(
        self,
        record: MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=f"{xpath}/gmd:CI_OnLineFunctionCode",
        )
        self.code_list_values = ["download", "information", "offlineAccess", "order", "search"]
        self.code_list = (
            "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/"
            "codelist/gmxCodelists.xml#CI_OnLineFunctionCode"
        )
        self.element = f"{{{self.ns.gmd}}}function"
        self.element_code = f"{{{self.ns.gmd}}}CI_OnLineFunctionCode"
        self.attribute = "function"


class MaintenanceInformation(MetadataRecordElement):
    def __init__(
        self,
        record: _MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=f"{xpath}/gmd:MD_MaintenanceInformation",
        )

    def make_config(self) -> dict:
        _ = {}

        maintenance_frequency = MaintenanceAndUpdateFrequency(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:maintenanceAndUpdateFrequency",
        )
        _maintenance_frequency = maintenance_frequency.make_config()
        if _maintenance_frequency != "":
            _["maintenance_frequency"] = _maintenance_frequency

        progress = MaintenanceProgress(
            record=self.record,
            attributes=self.attributes,
            xpath=f"{self.xpath}/gmd:maintenanceNote",
        )
        _progress = progress.make_config()
        if _progress != "":
            _["progress"] = _progress

        return _

    def make_element(self) -> None:
        maintenance_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}MD_MaintenanceInformation")

        if "maintenance_frequency" in self.element_attributes:
            maintenance_and_update_frequency = MaintenanceAndUpdateFrequency(
                record=self.record,
                attributes=self.attributes,
                parent_element=maintenance_element,
                element_attributes=self.element_attributes,
            )
            maintenance_and_update_frequency.make_element()

        if "progress" in self.element_attributes:
            maintenance_process = MaintenanceProgress(
                record=self.record,
                attributes=self.attributes,
                parent_element=maintenance_element,
                element_attributes=self.element_attributes,
            )
            maintenance_process.make_element()


class MaintenanceAndUpdateFrequency(CodeListElement):
    def __init__(
        self,
        record: MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=f"{xpath}/gmd:MD_MaintenanceFrequencyCode",
        )
        self.code_list_values = [
            "continual",
            "daily",
            "weekly",
            "fortnightly",
            "monthly",
            "quarterly",
            "biannually",
            "annually",
            "asNeeded",
            "irregular",
            "notPlanned",
            "unknown",
        ]
        self.code_list = (
            "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/"
            "codelist/gmxCodelists.xml#MD_MaintenanceFrequencyCode"
        )
        self.element = f"{{{self.ns.gmd}}}maintenanceAndUpdateFrequency"
        self.element_code = f"{{{self.ns.gmd}}}MD_MaintenanceFrequencyCode"
        self.attribute = "maintenance_frequency"


class MaintenanceProgress(CodeListElement):
    def __init__(
        self,
        record: MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=f"{xpath}/gmd:MD_ProgressCode",
        )
        self.code_list_values = [
            "completed",
            "historicalArchive",
            "obsolete",
            "onGoing",
            "planned",
            "required",
            "underDevelopment",
        ]
        self.code_list = (
            "http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/"
            "codelist/gmxCodelists.xml#MD_ProgressCode"
        )
        self.element = f"{{{self.ns.gmd}}}maintenanceNote"
        self.element_code = f"{{{self.ns.gmd}}}MD_ProgressCode"
        self.attribute = "progress"


class Citation(MetadataRecordElement):
    def __init__(
        self,
        record: _MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=f"{xpath}/gmd:CI_Citation",
        )

    def make_config(  # noqa: C901 see uk-pdc/metadata-infrastructure/metadata-library#175 for more information
        self,
    ) -> dict:
        _ = {}

        title_value = self.record.xpath(
            f"{self.xpath}/gmd:title/gco:CharacterString/text() | {self.xpath}/gmd:title/gmx:Anchor/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(title_value) == 1:
            if "title" not in _:
                _["title"] = {}
            _["title"]["value"] = title_value[0]

        title_href = self.record.xpath(f"{self.xpath}/gmd:title/gmx:Anchor/@xlink:href", namespaces=self.ns.nsmap())
        if len(title_href) == 1:
            if "title" not in _:
                _["title"] = {}
            _["title"]["href"] = title_href[0]

        _dates = {}
        dates_length = int(
            self.record.xpath(
                f"count({self.xpath}/gmd:date)",
                namespaces=self.ns.nsmap(),
            )
        )
        for date_index in range(1, dates_length + 1):
            date_ = Date(record=self.record, attributes=self.attributes, xpath=f"({self.xpath}/gmd:date)[{date_index}]")
            _date = date_.make_config()
            if bool(_date):
                _date_type = _date["date_type"]
                del _date["date_type"]
                _dates[_date_type] = _date
        if bool(_dates):
            _["dates"] = _dates

        edition_value = self.record.xpath(
            f"{self.xpath}/gmd:edition/gco:CharacterString/text()", namespaces=self.ns.nsmap()
        )
        if len(edition_value) == 1:
            _["edition"] = edition_value[0]

        _identifiers = []
        identifiers_length = int(
            self.record.xpath(
                f"count({self.xpath}/gmd:identifier)",
                namespaces=self.ns.nsmap(),
            )
        )
        for identifier_index in range(1, identifiers_length + 1):
            identifier = Identifier(
                record=self.record,
                attributes=self.attributes,
                xpath=f"({self.xpath}/gmd:identifier)[{identifier_index}]",
            )
            _identifier = identifier.make_config()
            if bool(_identifier):
                _identifiers.append(_identifier)
        if len(_identifiers) > 0:
            _["identifiers"] = _identifiers

        cited_responsible_party = ResponsibleParty(
            record=self.record, attributes=self.attributes, xpath=f"{self.xpath}/gmd:citedResponsibleParty"
        )
        _cited_responsible_party = cited_responsible_party.make_config()
        if bool(_cited_responsible_party):
            _["contact"] = _cited_responsible_party

        series = Series(record=self.record, attributes=self.attributes, xpath=f"{self.xpath}/gmd:series")
        _series = series.make_config()
        if bool(_series):
            _["series"] = _series

        other_citation_details_value = self.record.xpath(
            f"{self.xpath}/gmd:otherCitationDetails/gco:CharacterString/text()", namespaces=self.ns.nsmap()
        )
        if len(other_citation_details_value) == 1:
            _["other_citation_details"] = other_citation_details_value[0]

        return _

    def make_element(self) -> None:  # noqa: C901 see uk-pdc/metadata-infrastructure/metadata-library#175 for more information
        citation_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}CI_Citation")

        if "title" in self.element_attributes:
            title_element = SubElement(citation_element, f"{{{self.ns.gmd}}}title")
            if "href" in self.element_attributes["title"]:
                anchor = AnchorElement(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=title_element,
                    element_attributes=self.element_attributes["title"],
                    element_value=self.element_attributes["title"]["value"],
                )
                anchor.make_element()
            else:
                title_value = SubElement(title_element, f"{{{self.ns.gco}}}CharacterString")
                title_value.text = self.element_attributes["title"]["value"]

        if "dates" in self.element_attributes:
            for date_type, date_attributes in self.element_attributes["dates"].items():
                _date_attributes = deepcopy(date_attributes)
                _date_attributes["date_type"] = date_type
                citation_date = Date(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=citation_element,
                    element_attributes=_date_attributes,
                )
                citation_date.make_element()

        if "edition" in self.element_attributes:
            edition_element = SubElement(citation_element, f"{{{self.ns.gmd}}}edition")
            edition_value = SubElement(edition_element, f"{{{self.ns.gco}}}CharacterString")
            edition_value.text = str(self.element_attributes["edition"])

        if "identifiers" in self.element_attributes:
            for identifier_attributes in self.element_attributes["identifiers"]:
                identifier = Identifier(
                    record=self.record,
                    attributes=self.attributes,
                    parent_element=citation_element,
                    element_attributes=identifier_attributes,
                )
                identifier.make_element()

        if "contact" in self.element_attributes:
            citated_responsible_party_element = SubElement(citation_element, f"{{{self.ns.gmd}}}citedResponsibleParty")

            # Citations can only have a single contact so collapse roles array down to a single value
            _contact_element_attributes = self.element_attributes["contact"]
            if type(self.element_attributes["contact"]["role"]) is list:
                if len(self.element_attributes["contact"]["role"]) > 1:
                    msg = "Contacts can only have a single role. Citations can only have a single contact."
                    raise ValueError(msg)
                _contact_element_attributes = deepcopy(self.element_attributes["contact"])
                _contact_element_attributes["role"] = _contact_element_attributes["role"][0]

            responsible_party = ResponsibleParty(
                record=self.record,
                attributes=self.attributes,
                parent_element=citated_responsible_party_element,
                element_attributes=_contact_element_attributes,
            )
            responsible_party.make_element()

        if "series" in self.element_attributes:
            series = Series(
                record=self.record,
                attributes=self.attributes,
                parent_element=citation_element,
                element_attributes=self.element_attributes["series"],
            )
            series.make_element()

        if "other_citation_details" in self.element_attributes:
            other_citation_details_element = SubElement(citation_element, f"{{{self.ns.gmd}}}otherCitationDetails")
            other_citation_details_value = SubElement(
                other_citation_details_element, f"{{{self.ns.gco}}}CharacterString"
            )
            other_citation_details_value.text = str(self.element_attributes["other_citation_details"])


class Date(MetadataRecordElement):
    def __init__(
        self,
        record: _MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=f"{xpath}/gmd:CI_Date",
        )

    def make_config(self) -> dict:
        _ = {}

        date_value = self.record.xpath(
            f"{self.xpath}/gmd:date/gco:Date/text() | {self.xpath}/gmd:date/gco:DateTime/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(date_value) == 1:
            try:
                _ = decode_date_string(date_datetime=date_value[0])
            except ValueError:
                msg = "Date/datetime could not be parsed as an ISO date value"
                raise RuntimeError(msg) from None

        date_type = DateType(record=self.record, attributes=self.attributes, xpath=f"{self.xpath}/gmd:dateType")
        _date_type = date_type.make_config()
        _["date_type"] = _date_type

        return _

    def make_element(self) -> None:
        date_container_wrapper = SubElement(self.parent_element, f"{{{self.ns.gmd}}}date")
        date_container_element = SubElement(date_container_wrapper, f"{{{self.ns.gmd}}}CI_Date")

        date_element = SubElement(date_container_element, f"{{{self.ns.gmd}}}date")

        date_value_element = f"{{{self.ns.gco}}}Date"
        if type(self.element_attributes["date"]) is datetime:
            date_value_element = f"{{{self.ns.gco}}}DateTime"

        _date_precision = None
        if "date_precision" in self.element_attributes:
            _date_precision = self.element_attributes["date_precision"]
        date_value = SubElement(date_element, date_value_element)
        date_value.text = encode_date_string(
            date_datetime=self.element_attributes["date"], date_precision=_date_precision
        )

        date_type = DateType(
            record=self.record,
            attributes=self.attributes,
            parent_element=date_container_element,
            element_attributes=self.element_attributes,
        )
        date_type.make_element()


class DateType(CodeListElement):
    def __init__(
        self,
        record: MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=f"{xpath}/gmd:CI_DateTypeCode",
        )
        self.code_list_values = [
            "creation",
            "publication",
            "revision",
            "expiry",
            "lastUpdate",
            "lastRevision",
            "nextUpdate",
            "unavailable",
            "inForce",
            "adopted",
            "deprecated",
            "superseded",
            "validityBegins",
            "validityExpires",
            "released",
            "distribution",
        ]
        self.code_list = "https://standards.iso.org/iso/19115/resources/Codelists/cat/codelists.xml#CI_DateTypeCode"
        self.element = f"{{{self.ns.gmd}}}dateType"
        self.element_code = f"{{{self.ns.gmd}}}CI_DateTypeCode"
        self.attribute = "date_type"


class Identifier(MetadataRecordElement):
    def __init__(
        self,
        record: _MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
        identifier_container: str | None = f"{{{'http://www.isotc211.org/2005/gmd'}}}identifier",
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=f"{xpath}/gmd:RS_Identifier",
        )
        self.identifier_container = identifier_container

    def make_config(self) -> dict:
        _ = {}

        identifier_value = self.record.xpath(
            f"{self.xpath}/gmd:code/gco:CharacterString/text() | {self.xpath}/gmd:code/gmx:Anchor/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(identifier_value) == 1:
            _["identifier"] = identifier_value[0]

        identifier_href = self.record.xpath(f"{self.xpath}/gmd:code/gmx:Anchor/@xlink:href", namespaces=self.ns.nsmap())
        if len(identifier_href) == 1:
            _["href"] = identifier_href[0]

        identifier_namespace = self.record.xpath(
            f"{self.xpath}/gmd:codeSpace/gco:CharacterString/text()", namespaces=self.ns.nsmap()
        )
        if len(identifier_namespace) == 1:
            _["namespace"] = identifier_namespace[0]

        return _

    def make_element(self) -> None:
        identifier_container = SubElement(self.parent_element, self.identifier_container)
        identifier_wrapper = SubElement(identifier_container, f"{{{self.ns.gmd}}}RS_Identifier")
        identifier_element = SubElement(identifier_wrapper, f"{{{self.ns.gmd}}}code")

        if "href" in self.element_attributes:
            anchor = AnchorElement(
                record=self.record,
                attributes=self.attributes,
                parent_element=identifier_element,
                element_attributes=self.element_attributes,
                element_value=self.element_attributes["identifier"],
            )
            anchor.make_element()
        else:
            identifier_value = SubElement(identifier_element, f"{{{self.ns.gco}}}CharacterString")
            identifier_value.text = self.element_attributes["identifier"]

        if "namespace" in self.element_attributes:
            identifier_scheme_element = SubElement(identifier_wrapper, f"{{{self.ns.gmd}}}codeSpace")
            identifier_scheme_value = SubElement(identifier_scheme_element, f"{{{self.ns.gco}}}CharacterString")
            identifier_scheme_value.text = self.element_attributes["namespace"]


class AnchorElement(MetadataRecordElement):
    def __init__(
        self,
        record: MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        element_value: str | None = None,
    ):
        super().__init__(
            record=record, attributes=attributes, parent_element=parent_element, element_attributes=element_attributes
        )
        self.text = element_value

    def make_element(self) -> None:
        attributes = {}

        if "href" in self.element_attributes:
            attributes[f"{{{self.ns.xlink}}}href"] = self.element_attributes["href"]
            attributes[f"{{{self.ns.xlink}}}actuate"] = "onRequest"
        if "title" in self.element_attributes:
            attributes[f"{{{self.ns.xlink}}}title"] = self.element_attributes["title"]

        anchor = SubElement(self.parent_element, f"{{{self.ns.gmx}}}Anchor", attrib=attributes)
        if self.text is not None:
            anchor.text = self.text


class Format(MetadataRecordElement):
    def __init__(
        self,
        record: MetadataRecord,
        attributes: dict,
        parent_element: Element = None,
        element_attributes: dict | None = None,
        xpath: str | None = None,
    ):
        super().__init__(
            record=record,
            attributes=attributes,
            parent_element=parent_element,
            element_attributes=element_attributes,
            xpath=f"{xpath}/gmd:MD_Format",
        )

    def make_config(self) -> dict:
        _ = {}

        format_id = self.record.xpath(f"{self.xpath}/@id", namespaces=self.ns.nsmap())
        if len(format_id) == 1:
            _id: str = format_id[0].replace("bml-", "")
            _id = _id.replace("-fmt", "")
            _["_id"] = _id

        format_name = self.record.xpath(
            f"{self.xpath}/gmd:name/gco:CharacterString/text() | " f"{self.xpath}/gmd:name/gmx:Anchor/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(format_name) == 1:
            _["format"] = format_name[0]

        format_href = self.record.xpath(
            f"{self.xpath}/gmd:name/gmx:Anchor/@xlink:href",
            namespaces=self.ns.nsmap(),
        )
        if len(format_href) == 1:
            _["href"] = format_href[0]

        version_value = self.record.xpath(
            f"{self.xpath}/gmd:version/gco:CharacterString/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(version_value) == 1:
            _["version"] = version_value[0]

        amendment_number = self.record.xpath(
            f"{self.xpath}/gmd:amendmentNumber/gco:CharacterString/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(amendment_number) == 1:
            _["amendment_number"] = amendment_number[0]

        specification = self.record.xpath(
            f"{self.xpath}/gmd:specification/gco:CharacterString/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(specification) == 1:
            _["specification"] = specification[0]

        file_decompression_technique = self.record.xpath(
            f"{self.xpath}/gmd:fileDecompressionTechnique/gco:CharacterString/text()",
            namespaces=self.ns.nsmap(),
        )
        if len(file_decompression_technique) == 1:
            _["file_decompression_technique"] = file_decompression_technique[0]

        if list(_.keys()) == ["_id"]:
            _ = {}
        return _

    def make_element(self) -> None:
        if "_id" in self.element_attributes:
            format_element = SubElement(
                self.parent_element,
                f"{{{self.ns.gmd}}}MD_Format",
                attrib={"id": f"bml-{self.element_attributes['_id']}-fmt"},
            )
        else:
            format_element = SubElement(self.parent_element, f"{{{self.ns.gmd}}}MD_Format")

        format_name_element = SubElement(format_element, f"{{{self.ns.gmd}}}name")
        if "href" in self.element_attributes:
            anchor = AnchorElement(
                record=self.record,
                attributes=self.attributes,
                parent_element=format_name_element,
                element_attributes=self.element_attributes,
                element_value=self.element_attributes["format"],
            )
            anchor.make_element()
        else:
            format_name_value = SubElement(format_name_element, f"{{{self.ns.gco}}}CharacterString")
            format_name_value.text = self.element_attributes["format"]

        if "version" in self.element_attributes:
            version_element = SubElement(format_element, f"{{{self.ns.gmd}}}version")
            version_value = SubElement(version_element, f"{{{self.ns.gco}}}CharacterString")
            version_value.text = self.element_attributes["version"]
        else:
            SubElement(
                format_element,
                f"{{{self.ns.gmd}}}version",
                attrib={f"{{{self.ns.gco}}}nilReason": "missing"},
            )

        if "amendment_number" in self.element_attributes:
            amendment_number_element = SubElement(format_element, f"{{{self.ns.gmd}}}amendmentNumber")
            amendment_number_value = SubElement(amendment_number_element, f"{{{self.ns.gco}}}CharacterString")
            amendment_number_value.text = self.element_attributes["amendment_number"]

        if "specification" in self.element_attributes:
            specification_element = SubElement(format_element, f"{{{self.ns.gmd}}}specification")
            specification_element_value = SubElement(specification_element, f"{{{self.ns.gco}}}CharacterString")
            specification_element_value.text = self.element_attributes["specification"]

        if "file_decompression_technique" in self.element_attributes:
            file_decompression_technique_element = SubElement(
                format_element, f"{{{self.ns.gmd}}}fileDecompressionTechnique"
            )
            file_decompression_technique_value = SubElement(
                file_decompression_technique_element, f"{{{self.ns.gco}}}CharacterString"
            )
            file_decompression_technique_value.text = self.element_attributes["file_decompression_technique"]


class Series(MetadataRecordElement):
    """gmd:series."""

    def make_config(self) -> dict:
        """Decode to Python."""
        _ = {}

        name_value = self.record.xpath(
            f"{self.xpath}/gmd:CI_Series/gmd:name/gco:CharacterString/text()", namespaces=self.ns.nsmap()
        )
        if len(name_value) == 1:
            _["name"] = name_value[0]

        issue_value = self.record.xpath(
            f"{self.xpath}/gmd:CI_Series/gmd:issueIdentification/gco:CharacterString/text()", namespaces=self.ns.nsmap()
        )
        if len(issue_value) == 1:
            _["edition"] = issue_value[0]

        return _

    def make_element(self) -> None:
        """Encode as XML."""
        series_wrapper = SubElement(self.parent_element, f"{{{self.ns.gmd}}}series")
        series_element = SubElement(series_wrapper, f"{{{self.ns.gmd}}}CI_Series")

        if "name" in self.element_attributes:
            name_element = SubElement(series_element, f"{{{self.ns.gmd}}}name")
            name_value = SubElement(name_element, f"{{{self.ns.gco}}}CharacterString")
            name_value.text = self.element_attributes["name"]

        if "edition" in self.element_attributes:
            issue_element = SubElement(series_element, f"{{{self.ns.gmd}}}issueIdentification")
            issue_value = SubElement(issue_element, f"{{{self.ns.gco}}}CharacterString")
            issue_value.text = self.element_attributes["edition"]



from pathlib import Path
from tempfile import TemporaryDirectory
from zipfile import ZipFile

from lxml.etree import Element, fromstring


def generate_rtzp_archive(file: Path, rtz_name: str, rtz_document: str) -> None:
    """
    Generates RTZ record in a RTZP data container.

    A RTZP container is a zip archive containing a single RTZ file. This inner RTZ file must be named after the
    'route_name' config property in the record (e.g. if route_name is 'foo', the RTZ file must be named 'foo.rtz').

    :param file: path at which to create RTZP data container
    :param rtz_name: Name to use for the inner RTZ file
    :param rtz_document: Contents of the inner RTZ file as a string
    """
    with TemporaryDirectory() as tmpdirname:
        rtz_path = Path(tmpdirname).joinpath(f"{rtz_name}.rtz")
        with rtz_path.open(mode="w") as rtz_file:
            rtz_file.write(rtz_document)
        ZipFile(str(file), mode="w").write(filename=rtz_path, arcname=rtz_path.name)


def load_record_from_rtzp_archive(file: Path) -> Element:
    """
    Loads a RTZ record from a RTZP data container.

    :param file: path to RTZP data container
    :returns Parsed RTZ record as an XML element
    """
    with ZipFile(str(file)) as rtzp_file:
        return fromstring(rtzp_file.read(rtzp_file.namelist()[0]))  # noqa: S320 (see '`lxml` package (security)' README section)
